from imageatm.handlers.data_generator import TrainDataGenerator, ValDataGenerator
import pytest
import numpy as np

TEST_CONFIG = {"batch_size": 3, "n_classes": 2}

TEST_SAMPLES = [
    {"image_id": "helmet_1.jpg", "label": 0},
    {"image_id": "helmet_2.jpg", "label": 0},
    {"image_id": "helmet_3.jpg", "label": 0},
    {"image_id": "helmet_4.jpg", "label": 1},
    {"image_id": "helmet_5.jpg", "label": 1},
]


@pytest.fixture(autouse=True)
def common_patches(mocker):
    mocker.patch.object(TrainDataGenerator, 'data_generator')
    TrainDataGenerator.data_generator.return_value = 'X', 'y'

    mocker.patch.object(ValDataGenerator, 'data_generator')
    ValDataGenerator.data_generator.return_value = 'X', 'y'
    # Setting seed for np.random.shuffle in Train-mode
    np.random.seed(10247)


class TestTrainDataGenerator(object):

    generator = None

    def test__init(self, mocker):
        global generator
        generator = TrainDataGenerator(
            TEST_SAMPLES,
            'image_dir',
            TEST_CONFIG['batch_size'],
            TEST_CONFIG['n_classes'],
            'preprocess_input',
        )

        assert generator.samples == TEST_SAMPLES
        assert generator.img_dir == 'image_dir'
        assert generator.batch_size == 3
        assert generator.n_classes == 2
        assert generator.basenet_preprocess == 'preprocess_input'
        assert generator.img_load_dims == (256, 256)
        assert generator.img_crop_dims == (224, 224)
        assert generator.train is True

    def test__len(self):
        global generator
        x = generator.__len__()

        assert x == 2

    def test__get_item(self, mocker):
        global generator
        generator.__getitem__(1)

        generator.data_generator.assert_called_with(
            [{"image_id": "helmet_2.jpg", "label": 0}, {"image_id": "helmet_5.jpg", "label": 1}]
        )


class TestValDataGenerator(object):

    generator = None

    def test__init(self):
        global generator
        generator = ValDataGenerator(
            TEST_SAMPLES,
            'image_dir',
            TEST_CONFIG['batch_size'],
            TEST_CONFIG['n_classes'],
            'preprocess_input',
        )

        assert generator.samples == TEST_SAMPLES
        assert generator.img_dir == 'image_dir'
        assert generator.batch_size == 3
        assert generator.n_classes == 2
        assert generator.basenet_preprocess == 'preprocess_input'
        assert generator.img_load_dims == (224, 224)
        assert generator.train is False
        assert not hasattr(generator, 'img_crop_dims')

    def test__len(self):
        global generator
        assert generator.__len__() == 2

    def test__get_item(self, mocker):
        global generator
        generator.__getitem__(1)

        generator.data_generator.assert_called_with(
            [{"image_id": "helmet_4.jpg", "label": 1}, {"image_id": "helmet_5.jpg", "label": 1}]
        )



import os
import argparse
from imageatm.components import Training


def run_training(image_dir, job_dir, **kwargs):
    trainer = Training(image_dir, job_dir, **kwargs)
    trainer.train()


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-j', '--job-dir', help='Directory with job files.', required=True)
    parser.add_argument('-i', '--image-dir', help='Directory with image files.', required=True)
    parser.add_argument('--batch-size', help='Batch size.', required=False)
    parser.add_argument(
        '--epochs-train-dense', help='Number of epochs train only dense layer.', required=False
    )
    parser.add_argument(
        '--epochs-train-all', help='Number of epochs train all layers.', required=False
    )
    parser.add_argument('--learning-rate-dense', help='Learning rate dense layers.', required=False)
    parser.add_argument('--learning-rate-all', help='Learning rate all layers.', required=False)
    args = parser.parse_args()

    run_training(**{**args.__dict__, **os.environ})



from keras.models import load_model as load_model_keras
from tensorflow.python.client import device_lib
from multiprocessing import cpu_count
from keras.callbacks import Callback
from six import iteritems
import numpy as np


def load_model(model_path):
    return load_model_keras(model_path)


def use_multiprocessing():
    if __get_available_gpus():
        # if GPU is available, use all available CPUs for batch preprocessing
        use_multiprocessing = True
        num_workers = cpu_count()
    else:
        # device = 'CPU'
        use_multiprocessing = False
        num_workers = 1
    return use_multiprocessing, num_workers


def __get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']


class LoggingMetrics(Callback):
    """Callback for logging metrics at the end of each epoch.
    Default format looks like "Epoch: 3 - loss: 3.4123 - val_loss: 5.4321"
    # Arguments
        self.logger.debug_fcn: function for self.logger.debuging. default is self.logger.debug, which will self.logger.debug to standard output.
        format_epoch: format string for each epoch [default="Epoch: {} - {}"]
        format_keyvalue: format string for key value pairs [default="{}: {}"]
        format_separator= separator string between each pair [default=" - "]
    # Example
        ```python
        # Write to Python logging
        import logging
        model.fit(x, y, callbacks = [LoggingCallback(logging.info)]
        # Write to a file
        with open("log.txt", 'w') as f:
            def self.logger.debug_fcn(s):
                f.write(s)
                f.write("\n")
            model.fit(x, y, callbacks = [LoggingCallback(self.logger.debug_fcn)]
        ```
    """

    def __init__(
        self,
        logger,
        format_epoch="Epoch: {} - {}",
        format_keyvalue="{}: {:0.4f}",
        format_separator=" - ",
    ):
        Callback.__init__(self)
        self.logger = logger
        self.format_epoch = format_epoch
        self.format_keyvalue = format_keyvalue
        self.format_separator = format_separator

    def on_epoch_end(self, epoch, logs={}):
        values = self.format_separator.join(
            self.format_keyvalue.format(k, v) for k, v in iteritems(logs)
        )
        msg = self.format_epoch.format(epoch + 1, values)
        self.logger.debug(msg)


class LoggingModels(Callback):
    """
    """

    def __init__(
        self,
        filepath,
        logger,
        monitor='val_loss',
        verbose=0,
        save_best_only=False,
        save_weights_only=False,
        mode='auto',
        period=1,
    ):
        self.monitor = monitor
        self.verbose = verbose
        self.filepath = filepath
        self.save_best_only = save_best_only
        self.save_weights_only = save_weights_only
        self.period = period
        self.epochs_since_last_save = 0
        self.logger = logger

        if mode not in ['auto', 'min', 'max']:
            self.logger.warning(
                'ModelCheckpoint mode %s is unknown, ' 'fallback to auto mode.' % (mode),
                RuntimeWarning,
            )
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
            self.best = np.Inf
        elif mode == 'max':
            self.monitor_op = np.greater
            self.best = -np.Inf
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = np.greater
                self.best = -np.Inf
            else:
                self.monitor_op = np.less
                self.best = np.Inf

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = self.filepath.format(epoch=epoch + 1, **logs)
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    self.logger.warning(
                        'Can save best model only with %s available, ' 'skipping.' % (self.monitor),
                        RuntimeWarning,
                    )
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            self.logger.debug(
                                '\nEpoch %05d: %s improved from %0.5f to %0.5f,'
                                ' saving model to %s'
                                % (epoch + 1, self.monitor, self.best, current, filepath)
                            )
                        self.best = current
                        if self.save_weights_only:
                            self.model.save_weights(filepath, overwrite=True)
                        else:
                            self.model.save(filepath, overwrite=True)
                    else:
                        if self.verbose > 0:
                            self.logger.debug(
                                '\nEpoch %05d: %s did not improve from %0.5f'
                                % (epoch + 1, self.monitor, self.best)
                            )
            else:
                if self.verbose > 0:
                    self.logger.debug('\nEpoch %05d: saving model to %s' % (epoch + 1, filepath))
                if self.save_weights_only:
                    self.model.save_weights(filepath, overwrite=True)
                else:
                    self.model.save(filepath, overwrite=True)



import importlib
from keras.models import Model
from imageatm.handlers.image_classifier import ImageClassifier

import pytest

TEST_CONFIG = {
    "base_model_name": "MobileNet",
    "batch_size": 16,
    "decay_all": 0,
    "decay_dense": 0,
    "dropout_rate": 0.75,
    "epochs_train_all": 5,
    "epochs_train_dense": 5,
    "learning_rate_all": 0.0000003,
    "learning_rate_dense": 0.001,
    "multiprocessing_data_load": True,
    "num_workers_data_load": 8,
    "n_classes": 10,
    "loss": "categorical_crossentropy",
}


@pytest.fixture(autouse=True)
def common_patches(mocker):
    mocker.patch('keras.models.Model.compile')
    mocker.patch('keras.models.Model.fit_generator')
    mocker.patch('keras.models.Model.summary')


class TestImageClassifier(object):

    classifier = None

    def test__init(self):
        global classifier
        classifier = ImageClassifier(
            TEST_CONFIG['base_model_name'],
            TEST_CONFIG['n_classes'],
            TEST_CONFIG['learning_rate_dense'],
            TEST_CONFIG['dropout_rate'],
            TEST_CONFIG['loss'],
        )

        assert classifier.weights == 'imagenet'
        assert classifier.base_module is importlib.import_module('keras.applications.mobilenet')

    def test__build(self):
        global classifier
        classifier.build()

        assert isinstance(classifier.base_model, Model)
        assert isinstance(classifier.model, Model)

    def test__get_preprocess_input(self):
        global classifier
        classifier.get_preprocess_input()
        pass

    def test__get_base_layers(self):
        global classifier
        classifier.get_base_layers()
        pass

    def test__set_learning_rate(self):
        global classifier
        classifier.set_learning_rate(5)

        assert classifier.learning_rate == 5

    def test__compile(self):
        global classifier
        classifier.compile()

        Model.compile.assert_called()

    def test__summary(self):
        global classifier
        classifier.summary()

        Model.summary.assert_called()

    def test__fit_generator(self, mocker):
        global classifier
        classifier.fit_generator(
            generator='training_generator',
            validation_data='validation_generator',
            epochs=TEST_CONFIG['epochs_train_dense'],
            verbose=1,
            use_multiprocessing=TEST_CONFIG['multiprocessing_data_load'],
            workers=TEST_CONFIG['num_workers_data_load'],
            max_queue_size=30,
            callbacks=[],
        )

        Model.fit_generator.assert_called_once_with(
            callbacks=[],
            epochs=5,
            generator='training_generator',
            max_queue_size=30,
            use_multiprocessing=True,
            validation_data='validation_generator',
            verbose=1,
            workers=8,
        )



import os
from collections import Counter
from imageatm.components.prepare_data import DataPrep
from imageatm.handlers.utils import load_json

''' Files for test_valid_images'''
INVALID_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'image_invalid.jpg'
)

VALID_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'image_960x540.jpg'
)

''' Files for sample validation'''
TEST_STR_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_samples', 'test_str_labels.json'
)
TEST_INT_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_samples', 'test_samples_int.json'
)
TEST_FILE_STR2INT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_samples', 'test_int_labels.json'
)

TEST_IMG_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'test_img_dir'
)

TEST_STR_FILE_CORRUPTED = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    '../data/test_samples',
    'test_str_labels_corrupted.json',
)

dp = object.__new__(DataPrep)


def test_valid_images():
    expected = ['helmet_1.jpg', 'helmet_3.jpg', 'helmet_2.jpg', 'image_png.png']

    dp.image_dir = TEST_IMG_DIR
    dp.validate_images()

    assert dp.valid_image_ids == expected


def test_validate_samples():
    expected = load_json(TEST_STR_FILE)

    dp.image_dir = TEST_IMG_DIR
    dp.samples = load_json(TEST_STR_FILE)
    dp.min_class_size = 1
    dp.validate_samples()

    assert dp.samples == expected

    # exclude first 3 samples as they are corrupted
    expected = load_json(TEST_STR_FILE_CORRUPTED)[3:]

    dp.image_dir = TEST_IMG_DIR
    dp.samples = load_json(TEST_STR_FILE_CORRUPTED)
    dp.min_class_size = 1
    dp.validate_images()
    dp.validate_samples()

    assert dp.samples == expected


def test_get_class_mapping():
    dp.samples = load_json(TEST_STR_FILE)
    dp.get_class_mapping()
    expected = {0: 'left', 1: 'right'}
    assert dp.class_mapping == expected

    dp.samples = dp.samples[::-1]
    dp.get_class_mapping()
    expected = {0: 'left', 1: 'right'}
    assert dp.class_mapping == expected

    dp.samples = load_json(TEST_INT_FILE)
    print(dp.samples)
    dp.get_class_mapping()
    print(dp.class_mapping)
    expected = {0: 1, 1: 2}
    assert dp.class_mapping == expected

    dp.samples = dp.samples[::-1]
    dp.get_class_mapping()
    expected = {0: 1, 1: 2}
    assert dp.class_mapping == expected


def test_convert_label_to_int():
    dp.samples = load_json(TEST_STR_FILE)
    dp.convert_label_to_int()
    expected = load_json(TEST_FILE_STR2INT)

    assert dp.samples == expected


def test_split_samples_0():
    dp.samples = load_json(TEST_INT_FILE)
    dp.test_size = 0.2
    dp.val_size = 0.5

    dp.split_samples()

    assert len(dp.train_samples) == 12
    assert len(dp.val_samples) == 12
    assert len(dp.test_samples) == 6

    train_labels_count = Counter([i['label'] for i in dp.train_samples])
    val_labels_count = Counter([i['label'] for i in dp.val_samples])
    test_labels_count = Counter([i['label'] for i in dp.test_samples])

    assert train_labels_count[1] == 8
    assert train_labels_count[2] == 4

    assert val_labels_count[1] == 8
    assert val_labels_count[2] == 4

    assert test_labels_count[1] == 4
    assert test_labels_count[2] == 2


def test_split_samples_1():
    dp.samples = load_json(TEST_INT_FILE)
    dp.test_size = 0.2
    dp.val_size = 0.1

    dp.split_samples()

    assert len(dp.train_samples) == 21
    assert len(dp.val_samples) == 3
    assert len(dp.test_samples) == 6

    train_labels_count = Counter([i['label'] for i in dp.train_samples])
    val_labels_count = Counter([i['label'] for i in dp.val_samples])
    test_labels_count = Counter([i['label'] for i in dp.test_samples])

    assert train_labels_count[1] == 14
    assert train_labels_count[2] == 7

    assert val_labels_count[1] == 2
    assert val_labels_count[2] == 1

    assert test_labels_count[1] == 4
    assert test_labels_count[2] == 2



import logging
import os
import sys


def get_logger(name, job_dir):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    if not logger.handlers:
        # stream handler ensures that logging events are passed to stdout
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch_formatter = logging.Formatter('%(message)s')
        ch.setFormatter(ch_formatter)
        logger.addHandler(ch)

        # file handler ensures that logging events are passed to log file
        if not os.path.exists(job_dir):
            os.makedirs(job_dir)


        fh = logging.FileHandler(filename=os.path.join(job_dir, 'logs'))
        fh.setLevel(logging.DEBUG)
        fh_formatter = logging.Formatter(
            '%(asctime)s - %(module)s - %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S'
        )
        fh.setFormatter(fh_formatter)
        logger.addHandler(fh)

    return logger



from imageatm.components import AWS


def run_cloud(provider, tf_dir, region, instance_type, vpc_id, bucket, destroy, job_dir, **kwargs):
    if provider == 'aws':
        cloud = AWS(
            tf_dir=tf_dir,
            region=region,
            instance_type=instance_type,
            vpc_id=vpc_id,
            s3_bucket=bucket,
            job_dir=job_dir,
        )

    if destroy:
        cloud.destroy()

    else:
        cloud.init()
        cloud.apply()



import os
from imageatm.handlers.logger import get_logger
from imageatm.client.config import (
    validate_config,
    update_config,
    update_component_configs,
    config_set_image_dir,
    config_set_job_dir,
)


def pipeline(
    config,
    config_file,
    job_dir=None,
    image_dir=None,
    samples_file=None,
    provider=None,
    instance_type=None,
    region=None,
    vpc_id=None,
    bucket=None,
    tf_dir=None,
    train_cloud=None,
    destroy=None,
    resize=None,
    batch_size=None,
    learning_rate_dense=None,
    learning_rate_all=None,
    epochs_train_dense=None,
    epochs_train_all=None,
):
    '''Runs the entire pipeline based on config file.
    '''
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        samples_file=samples_file,
        provider=provider,
        instance_type=instance_type,
        region=region,
        vpc_id=vpc_id,
        bucket=bucket,
        tf_dir=tf_dir,
        train_cloud=train_cloud,
        destroy=destroy,
        resize=resize,
        batch_size=batch_size,
        learning_rate_dense=learning_rate_dense,
        learning_rate_all=learning_rate_all,
        epochs_train_dense=epochs_train_dense,
        epochs_train_all=epochs_train_all,
    )
    config = config_set_image_dir(config)
    config = config_set_job_dir(config)
    config = update_component_configs(config)

    validate_config(config, config.pipeline)

    logger = get_logger(__name__, config.job_dir)

    if 'data_prep' in config.pipeline:
        from imageatm.scripts import run_data_prep

        logger.info(
            '\n********************************\n'
            '******* Data preparation *******\n'
            '********************************'
        )

        run_data_prep(**config.data_prep)

        # update image_dir if images were resized
        if config.data_prep.get('resize', False):
            config.image_dir = os.path.join(config.image_dir, '_resized')
            config = update_component_configs(config)

    if 'train' in config.pipeline:
        logger.info(
            '\n********************************\n'
            '*********** Training ***********\n'
            '********************************'
        )

        if config.train.get('cloud'):
            from imageatm.scripts import run_training_cloud

            run_training_cloud(**{**config.cloud, **config.train})
        else:
            from imageatm.scripts import run_training

            run_training(**config.train)

    if 'evaluate' in config.pipeline:
        from imageatm.scripts import run_evaluation

        logger.info(
            '\n********************************\n'
            '********** Evaluation **********\n'
            '********************************'
        )

        run_evaluation(**config.evaluate)

    if 'cloud' in config.pipeline:
        from imageatm.scripts import run_cloud

        run_cloud(**config.cloud)


def dataprep(config, config_file, image_dir, samples_file, job_dir, resize):
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        samples_file=samples_file,
        resize=resize,
    )

    config.data_prep['run'] = True
    validate_config(config, ['data_prep'])

    from imageatm.scripts import run_data_prep

    run_data_prep(**config.data_prep)


def train(
    config,
    config_file=None,
    job_dir=None,
    image_dir=None,
    provider=None,
    instance_type=None,
    region=None,
    vpc_id=None,
    bucket=None,
    tf_dir=None,
    train_cloud=None,
    destroy=None,
    batch_size=None,
    learning_rate_dense=None,
    learning_rate_all=None,
    epochs_train_dense=None,
    epochs_train_all=None,
):
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        provider=provider,
        instance_type=instance_type,
        region=region,
        vpc_id=vpc_id,
        bucket=bucket,
        tf_dir=tf_dir,
        train_cloud=train_cloud,
        destroy=destroy,
        batch_size=batch_size,
        learning_rate_dense=learning_rate_dense,
        learning_rate_all=learning_rate_all,
        epochs_train_dense=epochs_train_dense,
        epochs_train_all=epochs_train_all,
    )

    config.train['run'] = True

    validate_config(config, ['train'])

    if config.train.get('cloud'):
        from imageatm.scripts import run_training_cloud

        run_training_cloud(**{**config.cloud, **config.train})
    else:
        from imageatm.scripts import run_training

        run_training(**config.train)


def evaluate(config, config_file=None, image_dir=None, job_dir=None):
    config = update_config(
        config=config, config_file=config_file, job_dir=job_dir, image_dir=image_dir
    )

    config.evaluate['run'] = True
    validate_config(config, ['evaluate'])

    from imageatm.scripts import run_evaluation

    run_evaluation(**config.evaluate)


def cloud(
    config,
    job_dir=None,
    config_file=None,
    provider=None,
    instance_type=None,
    region=None,
    vpc_id=None,
    bucket=None,
    tf_dir=None,
    train_cloud=None,
    destroy=None,
    no_destroy=None,
):
    config = update_config(
        config=config,
        job_dir=job_dir,
        config_file=config_file,
        provider=provider,
        instance_type=instance_type,
        region=region,
        vpc_id=vpc_id,
        bucket=bucket,
        tf_dir=tf_dir,
        train_cloud=train_cloud,
        destroy=destroy,
        no_destroy=no_destroy,
    )

    config.cloud['run'] = True
    validate_config(config, ['cloud'])

    from imageatm.scripts import run_cloud

    run_cloud(**config.cloud)



from imageatm.handlers.utils import load_yaml


# set job_dir if not already set by client option
def config_set_job_dir(config):
    if config.job_dir is None:
        if config.data_prep.get('job_dir') and config.data_prep.get('run'):
            config.job_dir = config.data_prep.get('job_dir')

        elif config.train.get('job_dir') and config.train.get('run'):
            config.job_dir = config.train.get('job_dir')

        elif config.evaluate.get('job_dir') and config.evaluate.get('run'):
            config.job_dir = config.evaluate.get('job_dir')

    return config


def config_set_image_dir(config):
    # set image_dir if not already set by client option
    if config.image_dir is None:
        if config.data_prep.get('image_dir') and config.data_prep.get('run'):
            config.image_dir = config.data_prep.get('image_dir')

        elif config.train.get('image_dir') and config.train.get('run'):
            config.image_dir = config.train.get('image_dir')

        elif config.evaluate.get('image_dir') and config.evaluate.get('run'):
            config.image_dir = config.evaluate.get('image_dir')

    return config


def update_component_configs(config):
    '''
    Populate central parameters to component configs
    '''
    if config.image_dir:
        config.data_prep['image_dir'] = config.image_dir
        config.train['image_dir'] = config.image_dir
        config.evaluate['image_dir'] = config.image_dir

    if config.job_dir:
        config.data_prep['job_dir'] = config.job_dir
        config.train['job_dir'] = config.job_dir
        config.evaluate['job_dir'] = config.job_dir
        config.cloud['job_dir'] = config.job_dir

    return config


def update_config(
    config,
    config_file=None,
    job_dir=None,
    image_dir=None,
    samples_file=None,
    provider=None,
    instance_type=None,
    region=None,
    vpc_id=None,
    bucket=None,
    tf_dir=None,
    train_cloud=None,
    destroy=None,
    no_destroy=None,
    resize=None,
    batch_size=None,
    learning_rate_dense=None,
    learning_rate_all=None,
    epochs_train_dense=None,
    epochs_train_all=None,
):
    # set defaults
    config.train['cloud'] = False
    config.data_prep['resize'] = False

    # load central config file
    if config_file:
        config_yml = load_yaml(config_file)

        # populate parameters from config file
        # parameters from config file overwrite defaults
        config.data_prep = {**config.data_prep, **config_yml.get('data_prep', {})}
        config.train = {**config.train, **config_yml.get('train', {})}
        config.evaluate = {**config.evaluate, **config_yml.get('evaluate', {})}
        config.cloud = {**config.cloud, **config_yml.get('cloud', {})}

        # set pipeline
        config.pipeline = [i for (i, j) in config_yml.items() if j.get('run')]

    # set options
    if job_dir:
        config.job_dir = job_dir

    if image_dir:
        config.image_dir = image_dir

    if samples_file:
        config.data_prep['samples_file'] = samples_file

    if provider:
        config.cloud['provider'] = provider

    if instance_type:
        config.cloud['instance_type'] = instance_type

    if region:
        config.cloud['region'] = region

    if vpc_id:
        config.cloud['vpc_id'] = vpc_id

    if bucket:
        config.cloud['bucket'] = bucket

    if tf_dir:
        config.cloud['tf_dir'] = tf_dir

    if train_cloud:
        config.train['cloud'] = train_cloud

    if destroy:
        config.cloud['destroy'] = destroy

    if no_destroy:
        config.cloud['destroy'] = False

    if resize:
        config.data_prep['resize'] = resize

    if batch_size:
        config.train['batch_size'] = batch_size

    if learning_rate_dense:
        config.train['learning_rate_dense'] = learning_rate_dense

    if learning_rate_all:
        config.train['learning_rate_all'] = learning_rate_all

    if epochs_train_dense:
        config.train['epochs_train_dense'] = epochs_train_dense

    if epochs_train_all:
        config.train['epochs_train_all'] = epochs_train_all

    config = update_component_configs(config)

    return config


def get_diff(name, config, required_keys, optional_keys):
    allowed_keys = required_keys + optional_keys

    msg = []
    # check that all required keys are in config keys
    diff = list(set(required_keys).difference(config.keys()))
    if diff:
        msg.append('{} config: missing required parameters [{}]\n'.format(name, ', '.join(diff)))

    # check that config keys are in allowed keys
    diff = list(set(config.keys()).difference(allowed_keys))
    if diff:
        msg.append(
            '{} config: [{}] not in allowed parameters [{}]\n'.format(
                name, ', '.join(diff), ', '.join(allowed_keys)
            )
        )

    return msg


def val_data_prep(config):
    required_keys = ['image_dir', 'job_dir', 'samples_file', 'run']
    optional_keys = ['resize']

    return get_diff('data_prep', config, required_keys, optional_keys)


def val_train(config):
    required_keys = ['image_dir', 'job_dir', 'cloud', 'run']
    optional_keys = [
        'batch_size',
        'learning_rate_dense',
        'learning_rate_all',
        'epochs_train_dense',
        'epochs_train_all',
    ]

    return get_diff('train', config, required_keys, optional_keys)


def val_evaluate(config):
    required_keys = ['image_dir', 'job_dir', 'run']
    optional_keys = []

    return get_diff('evaluate', config, required_keys, optional_keys)


def val_cloud(config):
    allowed_providers = ['aws']

    assert config.get('provider') is not None, 'Config error: cloud config: missing provider'

    provider = config.get('provider')
    assert (
        provider in allowed_providers
    ), 'Config error: cloud config: {} not in allowed providers [{}]'.format(
        provider, *allowed_providers
    )

    required_keys = {
        'aws': [
            'run',
            'job_dir',
            'tf_dir',
            'region',
            'vpc_id',
            'instance_type',
            'bucket',
            'destroy',
            'provider',
        ]
    }
    optional_keys = []

    return get_diff('cloud', config, required_keys[provider], optional_keys)


def validate_config(config, components):
    msgs = []

    for component in components:
        config_component = getattr(config, component)

        if component == 'data_prep':
            msgs += val_data_prep(config_component)

        if component == 'train':
            msgs += val_train(config_component)

        if component == 'evaluate':
            msgs += val_evaluate(config_component)

        if component == 'cloud':
            msgs += val_cloud(config_component)

    assert len(msgs) == 0, '\nConfig error:\n{}'.format(''.join(msgs))



from imageatm.components.training import Training
from imageatm.handlers.data_generator import TrainDataGenerator, ValDataGenerator
from imageatm.handlers.image_classifier import ImageClassifier
import pytest

## TODO: relative path instead of hard-coded
TEST_IMAGE_DIR = './tests/data/test_images'
TEST_JOB_DIR = './tests/data/test_train_job'


@pytest.fixture(autouse=True)
def common_patches(mocker):
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.__init__')
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.build')
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.compile')
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.fit_generator')
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.summary')
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.get_preprocess_input')
    mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.get_base_layers')
    ImageClassifier.__init__.return_value = None

    mocker.patch('imageatm.handlers.data_generator.TrainDataGenerator.__init__')
    mocker.patch('imageatm.handlers.data_generator.ValDataGenerator.__init__')
    TrainDataGenerator.__init__.return_value = None
    ValDataGenerator.__init__.return_value = None


class TestTraining(object):

    train = None

    def test__init(self):
        global train
        train = Training(image_dir=TEST_IMAGE_DIR, job_dir=TEST_JOB_DIR)

        assert train.n_classes == 2
        assert train.epochs_train_dense == 131

    def test__build_model(self):
        global train
        train.build_model()

        ImageClassifier.__init__.assert_called_once_with(
            'MobileNet', 2, 0.001, 0.75, 'categorical_crossentropy'
        )
        ImageClassifier.build.assert_called_once_with()

    def test__fit_model(self):
        global train
        train.fit_model()

        TrainDataGenerator.__init__.assert_called_once()
        ValDataGenerator.__init__.assert_called_once()
        ImageClassifier.get_preprocess_input.assert_called_with()

        # TODO: this tests are only rudimentary
        ImageClassifier.compile.assert_called()
        ImageClassifier.fit_generator.assert_called()
        ImageClassifier.get_base_layers.assert_called()



import os
import glob
import itertools
import numpy as np
import matplotlib.pyplot as plt
from imageatm.handlers.images import load_image
from imageatm.handlers.image_classifier import ImageClassifier
from imageatm.handlers.data_generator import ValDataGenerator
from imageatm.handlers.utils import load_json
from imageatm.handlers.keras_utils import use_multiprocessing, load_model
from imageatm.handlers.logger import get_logger
from sklearn.metrics import confusion_matrix, accuracy_score, classification_report
from vis.visualization import visualize_cam


BATCH_SIZE = 16
BASE_MODEL_NAME = 'MobileNet'

USE_MULTIPROCESSING, WORKERS = use_multiprocessing()


class Evaluation:
    """Evaluates saved model.

    Attributes:
        image_dir: path of image directory
        job_dir: path to job directory with samples
    """

    def __init__(self, image_dir, job_dir, batch_size=BATCH_SIZE, base_model_name=BASE_MODEL_NAME):
        """Inits Evaluation class with image_dir, job_dir, batch_size, base_model_name.

        Load test set and class mapping.
        Define variables: classes, n_classes, y_true, y_pred.
        Load the model with best available accuracy.
        """
        self.image_dir = os.path.abspath(image_dir)
        self.job_dir = os.path.abspath(job_dir)

        self.logger = get_logger(__name__, self.job_dir)
        self.samples_test = load_json(os.path.join(self.job_dir, 'test_samples.json'))
        self.class_mapping = load_json(os.path.join(self.job_dir, 'class_mapping.json'))
        self.n_classes = len(self.class_mapping.keys())
        self.classes = [str(self.class_mapping[str(i)]) for i in range(self.n_classes)]
        self.y_true = np.array([i['label'] for i in self.samples_test])
        self.batch_size = batch_size
        self.base_model_name = base_model_name

        self._set_plots()
        self._set_best_model()
        self._set_evaluation_dir()

    def _set_plots(self):
        ''' Checks whether ipython kernel is present.

        Plots will only be shown if in ipython, otherwise saved as files.
        '''
        try:
            __IPYTHON__
            self.show_plots = True
            self.save_plots = False
        except NameError:
            self.show_plots = False
            self.save_plots = True

    def _set_best_model(self):
        self.logger.info('\n****** Loading model ******\n')
        self.best_model, self.best_model_file = self._load_best_model()

    def _set_evaluation_dir(self):
        if self.save_plots:
            evaluation_dir_name = os.path.basename(self.best_model_file).split('.hdf5')[0]
            self.evaluation_dir = os.path.join(
                self.job_dir, 'evaluation_{}'.format(evaluation_dir_name)
            )
            if not os.path.exists(self.evaluation_dir):
                os.makedirs(self.evaluation_dir)

    def _load_best_model(self):
        model_files = glob.glob(os.path.join(self.job_dir, 'models', '*.hdf5'))
        max_acc_idx = np.argmax([os.path.basename(m).split('_')[3][:5] for m in model_files])
        best_model_file = model_files[max_acc_idx]
        best_model = load_model(best_model_file)
        self.logger.info('loaded {}\n'.format(best_model_file))
        return best_model, best_model_file

    def run(self):
        self.logger.info('\n****** Test set distribution ******\n')
        self.plot_test_set_distribution()

        self.classifier = ImageClassifier(
            base_model_name=self.base_model_name,
            n_classes=self.n_classes,
            weights=None,
            dropout_rate=None,
            learning_rate=None,
            loss=None,
        )
        self.classifier.model = self.best_model

        self.data_generator = ValDataGenerator(
            samples=self.samples_test,
            img_dir=self.image_dir,
            batch_size=self.batch_size,
            n_classes=self.n_classes,
            basenet_preprocess=self.classifier.get_preprocess_input(),
        )

        self.logger.info('\n****** Get predictions from test set ******\n')
        predictions_dist = self.classifier.predict_generator(
            data_generator=self.data_generator,
            workers=WORKERS,
            use_multiprocessing=USE_MULTIPROCESSING,
            verbose=1,
        )

        self.y_pred = np.argmax(predictions_dist, axis=1)
        self.accuracy = accuracy_score(y_true=self.y_true, y_pred=self.y_pred)

        self.logger.info(
            '\nModel achieves {}% accuracy on test set\n'.format(round(self.accuracy * 100, 2))
        )

        self.logger.info(self.classification_report())
        self.plot_confusion_matrix()

    def classification_report(self):
        return classification_report(
            y_true=self.y_true, y_pred=self.y_pred, target_names=self.classes
        )

    def confusion_matrix(self):
        return confusion_matrix(y_true=self.y_true, y_pred=self.y_pred)

    def plot_test_set_distribution(self):
        """Plots bars with number of samples for each label in test set."""
        counts = np.bincount(self.y_true)
        title = 'Number of images in test set: {}'.format(len(self.samples_test))
        index = np.arange(self.n_classes)
        title_fontsize = 16 if self.n_classes < 4 else 18
        text_fontsize = 12 if self.n_classes < 4 else 14

        plt.bar(index, counts)
        plt.xlabel('Label', fontsize=text_fontsize)
        plt.ylabel('Number of images', fontsize=text_fontsize)
        plt.xticks(index, self.classes, fontsize=text_fontsize, rotation=30)
        plt.title(title, fontsize=title_fontsize)

        # figsize = [min(15, self.n_classes * 2), 5]
        # plt.figure(figsize=figsize)
        plt.tight_layout()

        if self.save_plots:
            target_file = os.path.join(self.evaluation_dir, 'test_set_distribution.pdf')
            plt.savefig(target_file)
            self.logger.info('saved under {}'.format(target_file))

        if self.show_plots:
            plt.show()

    def plot_confusion_matrix(self):
        """Plots normalized confusion matrix."""
        cm = self.confusion_matrix()
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

        figsize = [min(15, self.n_classes * 3.5), min(15, self.n_classes * 3.5)]
        title_fontsize = 16 if self.n_classes < 4 else 18
        text_fontsize = 12 if self.n_classes < 4 else 14
        self.logger.info(figsize)
        plt.figure(figsize=figsize)
        plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
        plt.title('Confusion matrix', fontsize=title_fontsize)
        plt.colorbar()
        tick_marks = np.arange(self.n_classes)
        plt.xticks(tick_marks, self.classes, rotation=45, fontsize=text_fontsize)
        plt.yticks(tick_marks, self.classes, fontsize=text_fontsize)
        plt.ylabel('True label', fontsize=text_fontsize)
        plt.xlabel('Predicted label', fontsize=text_fontsize)

        thresh = cm.max() / 2.0
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            plt.text(
                j,
                i,
                '{:.2f}'.format(cm[i, j]),
                horizontalalignment='center',
                color='white' if cm[i, j] > thresh else 'black',
                fontsize=text_fontsize,
            )

        plt.tight_layout()

        if self.save_plots:
            target_file = os.path.join(self.evaluation_dir, 'confusion_matrix.pdf')
            plt.savefig(target_file)
            self.logger.info('saved under {}'.format(target_file))

        if self.show_plots:
            plt.show()

    # get all misclassified images:
    def get_correct_wrong_examples(self, label):
        """Gets correctly and wrongly predicted samples for a given label.

        Attributes:
            label: int or str (label for which the predictions should be considered)
        """
        correct = []
        wrong = []

        if type(label) == str:
            class_mapping_inv = {v: k for k, v in self.class_mapping.items()}
            label = int(class_mapping_inv[label])

        for i, sample in enumerate(self.samples_test):
            if self.y_true[i] == label:
                image_file = os.path.join(self.image_dir, sample['image_id'])
                if self.y_true[i] == self.y_pred[i]:
                    correct.append([i, load_image(image_file, target_size=(224, 224)), sample])
                else:
                    wrong.append([i, load_image(image_file, target_size=(224, 224)), sample])

        return correct, wrong

    # visualize misclassified images:
    def visualize_images(self, image_list, show_heatmap=False, n_plot=20):
        """Visualizes images in a sample list.

        Attributes:
            imgae_list: sample list
            show_heatmap: boolean (generates a gradient based class activation map (grad-CAM))
        """
        n_rows = min(n_plot, len(image_list))
        n_cols = 2 if show_heatmap else 1

        figsize = [5 * n_cols, 5 * n_rows]
        plt.figure(figsize=figsize)

        plot_count = 1
        for (i, img, sample) in image_list[:n_rows]:
            plt.subplot(n_rows, n_cols, plot_count)
            plt.imshow(img)
            plt.axis('off')
            plt.title(
                'true: {}, predicted: {}'.format(
                    self.class_mapping[str(self.y_true[i])], self.class_mapping[str(self.y_pred[i])]
                )
            )
            plot_count += 1

            if show_heatmap is True:
                heatmap = visualize_cam(
                    model=self.classifier.model,
                    layer_idx=89,
                    filter_indices=[self.y_pred[i]],
                    seed_input=self.classifier.get_preprocess_input()(
                        np.array(img).astype(np.float32)
                    ),
                )
                plt.subplot(n_rows, n_cols, plot_count)
                plt.imshow(img)
                plt.imshow(heatmap, alpha=0.7)
                plt.axis('off')
                plot_count += 1

        if self.save_plots:
            # TODO: pass name as argument
            target_file = os.path.join(self.evaluation_dir, 'misclassified_images.pdf')
            plt.savefig(target_file)
            self.logger.info('saved under {}'.format(target_file))

        if self.show_plots:
            plt.show()



import os
from imageatm.handlers import images as im


NON_EXISTENT_FILE_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'foo_bar.jpg'
)

INVALID_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'image_invalid.jpg'
)

TRUNCATED_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'truncated.jpg'
)

JPG_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'image_960x540.jpg'
)

PNG_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'image_png.png'
)

BMP_IMG_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '../data/test_images', 'image_bmp.bmp'
)


def test_validate_image():
    valid, error = im.validate_image(NON_EXISTENT_FILE_PATH)
    assert valid is False
    assert isinstance(error, FileNotFoundError)

    valid, error = im.validate_image(INVALID_IMG_PATH)
    assert valid is False
    assert isinstance(error, OSError)

    valid, error = im.validate_image(TRUNCATED_IMG_PATH)
    assert valid is False
    assert isinstance(error, OSError)

    valid, error = im.validate_image(JPG_IMG_PATH)
    assert valid is True
    assert error is None

    valid, error = im.validate_image(PNG_IMG_PATH)
    assert valid is True
    assert error is None

    valid, error = im.validate_image(BMP_IMG_PATH)
    assert valid is False
    assert error is None


def test_resize_image():
    img_960x540 = im.load_image(JPG_IMG_PATH)

    img_pp = im.resize_image(img_960x540, max_size=300, upscale=False)
    w, h = img_pp.size
    assert w == 533
    assert h == 300

    img_pp = im.resize_image(img_960x540, max_size=300, upscale=True)
    w, h = img_pp.size
    assert w == 533
    assert h == 300

    img_pp = im.resize_image(img_960x540, max_size=600, upscale=False)
    w, h = img_pp.size
    assert w == 960
    assert h == 540

    img_pp = im.resize_image(img_960x540, max_size=600, upscale=True)
    w, h = img_pp.size
    assert w == 1066
    assert h == 600



import os
import numpy as np
from imageatm.handlers.images import load_image, random_crop
from keras.utils import Sequence, to_categorical


class DataGenerator(Sequence):
    '''inherits from Keras Sequence base object, allows to use multiprocessing in .fit_generator'''

    def __init__(
        self, samples, img_dir, batch_size, n_classes, basenet_preprocess, img_load_dims, train
    ):
        self.samples = samples
        self.img_dir = img_dir
        self.batch_size = batch_size
        self.n_classes = n_classes
        self.basenet_preprocess = basenet_preprocess  # basenet specific preprocessing function
        self.img_load_dims = img_load_dims  # dimensions that images get resized into when loaded
        self.train = train
        self.on_epoch_end()  # call ensures that samples are shuffled in first epoch if shuffle is set to True

    def on_epoch_end(self):
        self.indexes = np.arange(len(self.samples))
        if self.train is True:
            np.random.shuffle(self.indexes)

    def __len__(self):
        return int(np.ceil(len(self.samples) / self.batch_size))  # number of batches per epoch

    def __getitem__(self, index):
        batch_indexes = self.indexes[
            index * self.batch_size : (index + 1) * self.batch_size
        ]  # get batch indexes
        batch_samples = [self.samples[i] for i in batch_indexes]  # get batch samples
        X, y = self.data_generator(batch_samples)
        return X, y

    def data_generator(self, batch_samples):
        # initialize images and labels tensors for faster processing
        dims = self.img_crop_dims if self.train == True else self.img_load_dims
        X = np.empty((len(batch_samples), *dims, 3))
        y = np.empty((len(batch_samples), self.n_classes))

        for i, sample in enumerate(batch_samples):
            # load and randomly augment image
            img_file = os.path.join(self.img_dir, sample['image_id'])
            img = np.asarray(load_image(img_file, self.img_load_dims))
            if self.train == True:
                img = random_crop(img, self.img_crop_dims)
            X[i,] = img

            # TODO: more efficient by preprocessing
            y[i,] = to_categorical([sample['label']], num_classes=self.n_classes)

        # apply basenet specific preprocessing
        # input is 4D numpy array of RGB values within [0, 255]
        X = self.basenet_preprocess(X)

        return X, y


class TrainDataGenerator(DataGenerator):
    def __init__(
        self,
        samples,
        img_dir,
        batch_size,
        n_classes,
        basenet_preprocess,
        img_load_dims=(256, 256),
        img_crop_dims=(224, 224),
        train=True,
    ):
        super(TrainDataGenerator, self).__init__(
            samples, img_dir, batch_size, n_classes, basenet_preprocess, img_load_dims, train
        )
        self.img_crop_dims = img_crop_dims  # dimensions that images get randomly cropped to


class ValDataGenerator(DataGenerator):
    def __init__(
        self,
        samples,
        img_dir,
        batch_size,
        n_classes,
        basenet_preprocess,
        img_load_dims=(224, 224),
        train=False,
    ):
        super(ValDataGenerator, self).__init__(
            samples, img_dir, batch_size, n_classes, basenet_preprocess, img_load_dims, train
        )



from imageatm.components.evaluation import Evaluation

import mock

TEST_IMAGE_DIR = './tests/data/test_images'
TEST_JOB_DIR = './tests/data/test_train_job'
TEST_BATCH_SIZE = 16
TEST_BASE_MODEL_NAME = 'MobileNet'

class TestEvaluation(object):

    eval = None

    def test__init(self):
        global eval
        eval = Evaluation(image_dir=TEST_IMAGE_DIR,
                          job_dir=TEST_JOB_DIR,
                          batch_size=TEST_BATCH_SIZE,
                          base_model_name=TEST_BASE_MODEL_NAME)

        assert eval.image_dir == TEST_IMAGE_DIR
        assert eval.job_dir == TEST_JOB_DIR
        assert len(eval.samples_test) == 4
        assert len(eval.class_mapping) == 2
        assert eval.n_classes == 2
        assert eval.classes == ['0', '1']
        assert eval.y_true[0] == 0
        assert eval.y_true[1] == 0
        assert eval.y_true[2] == 0
        assert eval.y_true[3] == 1
        assert eval.batch_size == 16
        assert eval.base_model_name == 'MobileNet'


    def test__run(self,mocker):
        mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.__init__', return_value=None)
        mocker.patch('imageatm.handlers.data_generator.ValDataGenerator.__init__', return_value=None)
        mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.get_preprocess_input')
        mocker.patch('imageatm.handlers.image_classifier.ImageClassifier.predict_generator')
        mocker.patch('imageatm.components.evaluation.np.argmax')
        mocker.patch('imageatm.components.evaluation.accuracy_score', return_value=0.4711)
        mp_plot_dist = mocker.patch('imageatm.components.evaluation.Evaluation.plot_test_set_distribution')
        mp_load = mocker.patch('imageatm.components.evaluation.Evaluation._load_best_model')
        mp_plot_cm = mocker.patch('imageatm.components.evaluation.Evaluation.plot_confusion_matrix')

        global eval
        eval.run()

        mp_plot_dist.assert_called()
        mp_load.assert_called()
        mp_plot_cm.assert_called()


    def test___load_best_model(self,mocker):
        mp = mocker.patch('imageatm.components.evaluation.load_model')

        global eval
        eval._load_best_model()

        mp.assert_called_with('./tests/data/test_train_job/models/model_mobilenet_15_0.375.hdf5')


    def test__plot_test_set_distribution(self, mocker):
        mock_plt_bar = mocker.patch('matplotlib.pyplot.bar')
        mock_plt_xlabel = mocker.patch('matplotlib.pyplot.xlabel')
        mock_plt_ylabel = mocker.patch('matplotlib.pyplot.ylabel')
        mock_plt_xticks = mocker.patch('matplotlib.pyplot.xticks')
        mock_plt_title = mocker.patch('matplotlib.pyplot.title')
        mock_plt_figure = mocker.patch('matplotlib.pyplot.figure')
        mock_plt_show = mocker.patch('matplotlib.pyplot.show')

        global eval
        eval.plot_test_set_distribution()

        mock_plt_bar.assert_called()
        mock_plt_xlabel.assert_called_with('Label', fontsize=12)
        mock_plt_ylabel.assert_called_with('Number of images', fontsize=12)
        mock_plt_xticks.assert_called()
        mock_plt_title.assert_called_with('Number of images in test set: 4', fontsize=16)
        mock_plt_figure.assert_called_with(figsize=[4, 5])
        mock_plt_show.assert_called()


    def test__plot_confusion_matrix(self, mocker):
        mock_plt_tight_layout = mocker.patch('matplotlib.pyplot.tight_layout')
        mock_plt_xlabel = mocker.patch('matplotlib.pyplot.xlabel')
        mock_plt_ylabel = mocker.patch('matplotlib.pyplot.ylabel')
        mock_plt_xticks = mocker.patch('matplotlib.pyplot.xticks')
        mock_plt_title = mocker.patch('matplotlib.pyplot.title')
        mock_plt_figure = mocker.patch('matplotlib.pyplot.figure')
        mock_plt_imshow = mocker.patch('matplotlib.pyplot.imshow')
        mock_plt_colorbar = mocker.patch('matplotlib.pyplot.colorbar')
        mock_plt_xticks = mocker.patch('matplotlib.pyplot.xticks')
        mock_plt_yticks = mocker.patch('matplotlib.pyplot.yticks')
        mock_plt_show = mocker.patch('matplotlib.pyplot.show')

        global eval
        eval.y_pred = [1, 0, 0, 0]
        eval.plot_confusion_matrix()

        mock_plt_figure.assert_called()
        mock_plt_imshow.assert_called()
        mock_plt_title.assert_called_with('Confusion matrix', fontsize=16)
        mock_plt_imshow.assert_called()
        mock_plt_colorbar.assert_called()
        mock_plt_xticks.assert_called()
        mock_plt_yticks.assert_called()
        mock_plt_xlabel.assert_called_with('Predicted label', fontsize=12)
        mock_plt_ylabel.assert_called_with('True label', fontsize=12)
        mock_plt_tight_layout.assert_called()
        mock_plt_show.assert_called()
        pass


    def test__get_correct_wrong_examples(self):

        # eval.y_true = [0, 0, 0, 1]
        def paramized_test_correct_wrong_function(y_pred, label, num_correct, num_wrong):
            global eval
            eval.y_pred = y_pred
            correct, wrong = eval.get_correct_wrong_examples(label)
            assert len(correct) == num_correct
            assert len(wrong) == num_wrong

        paramized_test_correct_wrong_function([0, 0, 0, 1], 0, 3, 0)
        paramized_test_correct_wrong_function([0, 0, 0, 1], 1, 1, 0)
        paramized_test_correct_wrong_function([1, 0, 0, 0], 0, 2, 1)
        paramized_test_correct_wrong_function([1, 0, 0, 0], 1, 0, 1)


    def test__visualize_images(selfs):
        pass




from imageatm.components import DataPrep


def run_data_prep(image_dir, samples_file, job_dir, resize, **kwargs):

    dp = DataPrep(job_dir=job_dir, image_dir=image_dir, samples_file=samples_file)
    dp.prepare_data(resize=resize)



import os
import numpy as np
import tensorflow as tf
from imageatm.handlers.data_generator import TrainDataGenerator, ValDataGenerator
from imageatm.handlers.image_classifier import ImageClassifier
from imageatm.handlers.utils import load_json
from imageatm.handlers.keras_utils import (
    use_multiprocessing,
    LoggingMetrics,
    LoggingModels,
)
from imageatm.handlers.logger import get_logger
from keras import backend as K
from keras.callbacks import ReduceLROnPlateau

tf.logging.set_verbosity(tf.logging.ERROR)

LEARNING_RATE_DENSE = 0.001
LEARNING_RATE_ALL = 0.0001
BATCH_SIZE = 64
DROPOUT_RATE = 0.75
BASE_MODEL_NAME = 'MobileNet'
LOSS = 'categorical_crossentropy'


class Training:
    def __init__(
        self,
        image_dir,
        job_dir,
        epochs_train_dense=None,  # if None then will be set based on number of samples
        epochs_train_all=None,  # if None then will be set based on number of samples
        learning_rate_dense=LEARNING_RATE_DENSE,
        learning_rate_all=LEARNING_RATE_ALL,
        batch_size=BATCH_SIZE,
        dropout_rate=DROPOUT_RATE,
        base_model_name=BASE_MODEL_NAME,
        loss=LOSS,
        **kwargs
    ):

        self.image_dir = os.path.abspath(image_dir)
        self.job_dir = os.path.abspath(job_dir)

        self.logger = get_logger(__name__, self.job_dir)
        self.samples_train = load_json(os.path.join(self.job_dir, 'train_samples.json'))
        self.samples_val = load_json(os.path.join(self.job_dir, 'val_samples.json'))
        self.class_mapping = load_json(os.path.join(self.job_dir, 'class_mapping.json'))
        self.n_classes = len(self.class_mapping.keys())

        self.epochs_train_dense = epochs_train_dense
        self.epochs_train_all = epochs_train_all
        self.learning_rate_dense = float(learning_rate_dense)
        self.learning_rate_all = float(learning_rate_all)
        self.batch_size = int(batch_size)
        self.dropout_rate = float(dropout_rate)
        self.base_model_name = base_model_name
        self.loss = loss
        self.use_multiprocessing, self.workers = use_multiprocessing()
        self.calc_epochs_train()

    def build_model(self):
        self.classifier = ImageClassifier(
            self.base_model_name,
            self.n_classes,
            self.learning_rate_dense,
            self.dropout_rate,
            self.loss,
        )
        self.classifier.build()

    def fit_model(self):
        training_generator = TrainDataGenerator(
            self.samples_train,
            self.image_dir,
            self.batch_size,
            self.n_classes,
            self.classifier.get_preprocess_input(),
        )

        validation_generator = ValDataGenerator(
            self.samples_val,
            self.image_dir,
            self.batch_size,
            self.n_classes,
            self.classifier.get_preprocess_input(),
        )

        # TODO: initialize callbacks TensorBoardBatch and ModelCheckpoint
        # tensorboard = TensorBoardBatch(log_dir=os.path.join(job_dir, 'logs'))

        model_save_name = (
            'model_' + self.base_model_name.lower() + '_{epoch:02d}_{val_acc:.3f}.hdf5'
        )
        model_dir = os.path.join(self.job_dir, 'models')
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)

        logging_metrics = LoggingMetrics(logger=self.logger)
        logging_models = LoggingModels(
            logger=self.logger,
            filepath=os.path.join(model_dir, model_save_name),
            monitor='val_acc',
            verbose=1,
            save_best_only=True,
            save_weights_only=False,
        )

        def __train_dense_layers():
            self.logger.info('\n****** Train dense layers ******\n')

            # freeze convolutional layers in base net
            for layer in self.classifier.get_base_layers():
                layer.trainable = False

            self.classifier.compile()
            # self.classifier.summary()

            self.classifier.fit_generator(
                generator=training_generator,
                validation_data=validation_generator,
                epochs=self.epochs_train_dense,
                verbose=1,
                use_multiprocessing=self.use_multiprocessing,
                workers=self.workers,
                max_queue_size=30,
                callbacks=[logging_metrics, logging_models],
            )

        def __train_all_layers():
            self.logger.info('\n****** Train all layers ******\n')

            min_lr = self.learning_rate_all / 10
            reduce_lr = ReduceLROnPlateau(
                monitor='val_acc', factor=0.2, patience=5, min_lr=min_lr, verbose=1
            )

            # unfreeze all layers
            for layer in self.classifier.get_base_layers():
                layer.trainable = True

            self.classifier.set_learning_rate(self.learning_rate_all)

            self.classifier.compile()
            # self.classifier.summary()

            self.classifier.fit_generator(
                generator=training_generator,
                validation_data=validation_generator,
                epochs=self.epochs_train_dense + self.epochs_train_all,
                initial_epoch=self.epochs_train_dense,
                verbose=1,
                use_multiprocessing=self.use_multiprocessing,
                workers=self.workers,
                max_queue_size=30,
                callbacks=[logging_metrics, logging_models, reduce_lr],
            )

        __train_dense_layers()
        __train_all_layers()

        K.clear_session()

    def calc_epochs_train(self):
        '''
        we assume a simplified logic behind the interpolation: 100/2^(log(x)-1)
        '''
        x = len(self.samples_train)

        if not self.epochs_train_dense:
            self.epochs_train_dense = int(100 / (2 ** (np.log10(x) - 1)))
        else:
            self.epochs_train_dense = int(self.epochs_train_dense)

        if not self.epochs_train_all:
            self.epochs_train_all = int(100 / (2 ** (np.log10(x) - 1)))
        else:
            self.epochs_train_all = int(self.epochs_train_all)

    def train(self):
        self.build_model()
        self.fit_model()



import os
from collections import Counter
from imageatm.handlers.images import resize_image_mp, validate_image
from imageatm.handlers.utils import load_json, save_json, parallelise
from imageatm.handlers.logger import get_logger
from sklearn.model_selection import train_test_split


class DataPrep:
    """Prepares data for training.

    Attributes:
        image_dir: path of image directory
        job_dir: path to job directory with samples
    """

    def __init__(
        self,
        job_dir,
        image_dir,
        samples_file,
        min_class_size=2,
        test_size=0.2,
        val_size=0.1,
    ):
        """
        Inits data preparation object.

        Load samples file.
        Initialize variables for further operations:
        valid_image_ids, class_mapping, train_samples, val_samples, test_samples
        """
        self.job_dir = job_dir

        self.logger = get_logger(__name__, self.job_dir)
        self.image_dir = image_dir
        self.samples_file = samples_file
        self.min_class_size = min_class_size
        self.test_size = test_size
        self.val_size = val_size

        self.valid_image_ids = None
        self.class_mapping = None
        self.train_samples = None
        self.val_samples = None
        self.test_samples = None

        self.samples = load_json(self.samples_file)

    def prepare_data(self, resize=False):
        """Executes all steps of data preparation."""
        self.validate_images()
        self.validate_samples()
        self.get_class_mapping()
        self.convert_label_to_int()
        self.split_samples()

        if resize:
            self.resize_images()

        self.save_files()

    def get_counter(self, list_to_count, print_count=False):
        counter = Counter(list_to_count)
        total = len(list_to_count)

        if print_count:
            for key, val in sorted(counter.items()):
                self.logger.info(
                    '{}: {} ({}%)'.format(key, val, round(val * 100 / total, 1))
                )

        return counter

    def get_samples_count(self, print_count=True):
        """Retrieves samples count for each label from loaded samples."""
        self.samples_count = self.get_counter(
            [i['label'] for i in self.samples], print_count
        )

    def validate_images(self):
        """Checks if all files in image_dir are valid images.

        Checks if files are images with extention 'JPEG' or 'PNG' and if they are not truncated.
        self.logger.infos filenames that didn't pass the validationself.

        Sets:
            self.valid_image_ids: a list of valid image
        """
        self.logger.info('\n****** Running image validation ******\n')

        # validate images, use multiprocessing
        files = [os.path.join(self.image_dir, i) for i in os.listdir(self.image_dir)]
        results = parallelise(validate_image, files)

        valid_image_files = [j for i, j in enumerate(files) if results[i][0]]
        self.valid_image_ids = [os.path.basename(i) for i in valid_image_files]

        # return list of invalid images to user and save them if there are more than 10
        invalid_image_files = [
            (j, results[i][1]) for i, j in enumerate(files) if not results[i][0]
        ]

        if invalid_image_files:
            self.logger.info('The following files are not valid image files:')
            for file_name, error_msg in invalid_image_files[:10]:
                self.logger.info('- {} ({})'.format(file_name, error_msg))
            if len(invalid_image_files) > 10:
                save_json(
                    invalid_image_files,
                    os.path.join(self.job_dir, 'invalid_image_files.json'),
                )
                self.logger.info(
                    (
                        'NOTE: More than 10 files were identified as invalid image files.\n'
                        'The full list of those files has been saved here:\n{}'.format()
                    )
                )

    @staticmethod
    def _validate_sample(sample, valid_image_ids):
        return all(
            [
                ('label' in sample),
                ('image_id' in sample),
                (sample.get('image_id') in valid_image_ids),
            ]
        )

    def validate_samples(self):
        """Validates the samples.

        Compares samples with valid image ids.
        Checks if :
            - samples have 'label' and 'image_id' keys
            - there is more than 1 class
            - there are enough samples in each class

        Reassigns self.samples to valid samples.
        self.logger.infos:
            - samples that didn't pass the validationself
            - distribution of valid samples for each label
        """
        # images have to be validated before samples can be validated
        if not self.valid_image_ids:
            self.validate_images()

        self.logger.info('\n****** Running samples validation ******\n')

        valid_image_ids = set(self.valid_image_ids)  # convert to set to optimise lookup

        # exclude samples with invalid image or invalid keys
        valid_samples = []
        self.invalid_samples = []
        for sample in self.samples:
            if self._validate_sample(sample, valid_image_ids):
                valid_samples.append(sample)
            else:
                self.invalid_samples.append(sample)

        # replace self.samples with valid samples
        assert valid_samples, 'Program ended. No valid samples found.'
        self.samples = valid_samples

        # self.logger.infos invalid images to user and saves them in json if there are more than 10
        if self.invalid_samples:
            self.logger.info('The following samples were dropped:')
            for sample in self.invalid_samples[:10]:
                self.logger.info('- {}'.format(sample))

            if len(self.invalid_samples) > 10:
                self.logger.info(
                    (
                        'NOTE: {} samples were identified as invalid.\n'
                        'The full list of invalid samples will be saved in job dir.\n'.format(
                            len(self.invalid_samples)
                        )
                    )
                )

        self.logger.info('Class distribution after validation:')
        self.get_samples_count()

        # check if each class has sufficient samples
        warnings_count = 0
        for key, val in self.samples_count.items():
            if val < self.min_class_size:
                warnings_count += 1
                self.logger.info('Not enough samples for label {}'.format(key))

        assert warnings_count == 0, 'Program ended. Collect more samples.'
        assert (
            len(self.samples_count) > 1
        ), 'Program ended. Only one label in the dataset.'

    def get_class_mapping(self, print_mapping=True):
        self.get_samples_count(print_count=False)

        self.class_mapping = {}
        labels = list(self.samples_count.keys())
        labels.sort()

        for i, j in enumerate(labels):
            self.class_mapping[i] = j

        if print_mapping:
            self.logger.info('Class mapping:\n{}'.format(self.class_mapping))

    def convert_label_to_int(self):
        self.get_class_mapping(print_mapping=False)

        class_mapping_inv = {v: k for k, v in self.class_mapping.items()}

        samples_int = [
            {
                'label': class_mapping_inv[sample['label']],
                'image_id': sample['image_id'],
            }
            for sample in self.samples
        ]

        # replace self.samples with samples int
        self.samples = samples_int

    def split_samples(self):
        """Produces stratified train, val, test sets.

        The test proportion is applied first, then val proportion,
        i.e. if test_size=0.2 and val_size=0.1,
        then train_size=0.72 ((1-0.2)*0.9).

        Sets:
            self.train_samples, self.val_samples, self.test_samples
        """
        self.logger.info('\n****** Creating train/val/test sets ******\n')

        train_size = (1 - self.test_size) * (1 - self.val_size)

        self.logger.info(
            'Split distribution: train: {:.2f}, val: {}, test: {:}\n'.format(
                train_size, self.val_size, self.test_size
            )
        )

        labels = [
            i['label'] for i in self.samples
        ]  # need label list for stratification

        train_samples, self.test_samples, train_labels, _ = train_test_split(
            self.samples,
            labels,
            test_size=self.test_size,
            shuffle=True,
            random_state=10207,
            stratify=labels,
        )

        self.train_samples, self.val_samples, _, _ = train_test_split(
            train_samples,
            train_labels,
            test_size=self.val_size,
            shuffle=True,
            random_state=10207,
            stratify=train_labels,
        )

        self.logger.info('Train set:')
        self.get_counter([i['label'] for i in self.train_samples], print_count=True)

        self.logger.info('Val set:')
        self.get_counter([i['label'] for i in self.val_samples], print_count=True)

        self.logger.info('Test set:')
        self.get_counter([i['label'] for i in self.test_samples], print_count=True)

    def resize_images(self, resize_image_mp=resize_image_mp):
        self.logger.info('\n****** Resizing images ******\n')
        new_image_dir = os.path.join(self.image_dir, '_resized')
        if not os.path.exists(new_image_dir):
            os.makedirs(new_image_dir)

        args = [(self.image_dir, new_image_dir, i['image_id']) for i in self.samples]
        parallelise(resize_image_mp, args)
        self.logger.info(
            'Stored {} resized images under {}'.format(len(self.samples), new_image_dir)
        )

        self.image_dir = new_image_dir
        self.logger.info('Changed image directory to {}'.format(self.image_dir))

    def save_files(self, dir=None):
        if dir:
            self.job_dir = dir
            if not os.path.exists(dir):
                os.makedirs(dir)

        else:
            self.job_dir = os.path.abspath(self.job_dir)

        if not self.train_samples:
            self.split_samples()

        if not self.class_mapping:
            self.get_class_mapping()

        save_json(self.train_samples, os.path.join(self.job_dir, 'train_samples.json'))
        save_json(self.val_samples, os.path.join(self.job_dir, 'val_samples.json'))
        save_json(self.test_samples, os.path.join(self.job_dir, 'test_samples.json'))
        save_json(self.class_mapping, os.path.join(self.job_dir, 'class_mapping.json'))

        if len(self.invalid_samples) > 10:
            save_json(
                self.invalid_samples, os.path.join(self.job_dir, 'invalid_samples.json')
            )
            self.logger.info(
                (
                    'NOTE: More than 10 samples were identified as invalid.\n'
                    'The full list of invalid samples has been saved here:\n{}'.format(
                        os.path.join(self.job_dir, 'invalid_samples.json')
                    )
                )
            )
        self.logger.info('\n')



from imageatm.handlers.keras_utils import use_multiprocessing

import mock

class TestKerasUtils(object):

    @mock.patch('imageatm.handlers.keras_utils.__get_available_gpus', return_value=[])
    def test__use_multiprocessing_false(self, mock):
        use_multi, num_worker = use_multiprocessing()

        assert use_multi == False
        assert num_worker == 1

    @mock.patch('imageatm.handlers.keras_utils.cpu_count', return_value=4711)
    @mock.patch('imageatm.handlers.keras_utils.__get_available_gpus', return_value=['foo', 'bar'])
    def test__use_multiprocessing_true(self, mock1, mock2):
        use_multi, num_worker = use_multiprocessing()

        assert use_multi == True
        assert num_worker == 4711



import argparse
from imageatm.components import AWS


def run_training_cloud(
    image_dir, job_dir, provider, tf_dir, region, instance_type, vpc_id, bucket, destroy, **kwargs
):
    if provider == 'aws':
        cloud = AWS(
            tf_dir=tf_dir,
            region=region,
            instance_type=instance_type,
            vpc_id=vpc_id,
            s3_bucket=bucket,
            job_dir=job_dir,
        )

    cloud.init()
    cloud.apply()
    cloud.train(image_dir=image_dir, job_dir=job_dir, **kwargs)

    if destroy:
        cloud.destroy()


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--job-dir', help='Directory with job files.', required=True)
    parser.add_argument('--image-dir', help='Directory with image files.', required=True)
    parser.add_argument('--provider', help='Cloud provider [aws].', required=True)
    parser.add_argument('--tf-dir', help='Directory with Terraform files.', required=True)
    parser.add_argument('--region', help='Cloud region.', required=True)
    parser.add_argument('--instance-type', help='Cloud instance type.', required=True)
    parser.add_argument('--vpc-id', help='Cloud vpc id [aws].', required=True)
    parser.add_argument('--bucket', help='Cloud storage.', required=True)
    parser.add_argument('--batch-size', help='Batch size.', required=False)
    parser.add_argument(
        '--destroy',
        help='Whether instance shopuld be terminated after training.',
        action='store_true',
        default=False,
        required=True,
    )
    parser.add_argument(
        '--epochs-train-dense', help='Number of epochs train only dense layer.', required=False
    )
    parser.add_argument(
        '--epochs-train-all', help='Number of epochs train all layers.', required=False
    )
    parser.add_argument('--learning-rate-dense', help='Learning rate dense layers.', required=False)
    parser.add_argument('--learning-rate-all', help='Learning rate all layers.', required=False)

    args = parser.parse_args()

    run_training_cloud(**args.__dict__)



import importlib
from keras.models import Model
from keras.layers import Dropout, Dense
from keras.optimizers import Adam


class ImageClassifier:
    def __init__(
        self, base_model_name, n_classes, learning_rate, dropout_rate, loss, weights='imagenet'
    ):
        self.n_classes = n_classes
        self.base_model_name = base_model_name
        self.learning_rate = learning_rate
        self.dropout_rate = dropout_rate
        self.loss = loss
        self.weights = weights
        self.__load_base_module()

    def __load_base_module(self):
        # import Keras base model module
        if self.base_model_name == 'InceptionV3':
            self.base_module = importlib.import_module('keras.applications.inception_v3')
        elif self.base_model_name == 'InceptionResNetV2':
            self.base_module = importlib.import_module('keras.applications.inception_resnet_v2')
        else:
            self.base_module = importlib.import_module(
                'keras.applications.' + self.base_model_name.lower()
            )

    def get_base_layers(self):
        return self.base_model.layers

    def get_preprocess_input(self):
        return self.base_module.preprocess_input

    def set_learning_rate(self, learning_rate):
        self.learning_rate = learning_rate

    def build(self):
        # get base model class
        BaseCnn = getattr(self.base_module, self.base_model_name)

        # load pre-trained model
        self.base_model = BaseCnn(
            input_shape=(224, 224, 3), weights=self.weights, include_top=False, pooling='avg'
        )

        # add dropout and dense layer
        x = Dropout(self.dropout_rate)(self.base_model.output)
        x = Dense(units=self.n_classes, activation='softmax')(x)

        self.model = Model(self.base_model.inputs, x)

    def compile(self):
        self.model.compile(
            optimizer=Adam(lr=self.learning_rate), loss=self.loss, metrics=['accuracy']
        )

    def fit_generator(self, **kwargs):
        self.model.fit_generator(**kwargs)

    def predict_generator(self, data_generator, **kwargs):
        return self.model.predict_generator(data_generator, **kwargs)

    def summary(self):
        self.model.summary()



from setuptools import setup, find_packages

setup(
    name='imageatm',
    version='0.0.1',
    author='Christopher Lennan, Malgorzata Adamczyk, Gunar Maiwald, Dat Tran',
    author_email='christopher.lennan@idealo.de, malgorzata.adamczyk@idealo.de, gunar.maiwald@idealo.de, dat.tran@idealo.de',
    description='Image classification for everyone',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    install_requires=['Keras>=2.2.4',
                      'keras-vis>=0.4.1',
                      'tensorflow>=1.12.0',
                      'awscli',
                      'Click',
                      'h5py',
                      'matplotlib',
                      'Pillow',
                      'python-dateutil',
                      'scikit-learn',
                      'scipy',
                      'tqdm',
                    ],
    extras_require={
        'tests': ['pytest',
                 'pytest-cov',
                 'pytest-mock',
                 'mock',
                 ]
    },
    classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    packages=find_packages(),
    entry_points={'console_scripts': ['image-atm=imageatm.client.client:cli']},
)


import os
import time
from imageatm.handlers.utils import run_cmd
from imageatm.handlers.logger import get_logger


class AWS:
    def __init__(self, tf_dir, region, instance_type, vpc_id, s3_bucket, job_dir):
        self.tf_dir = tf_dir
        self.region = region
        self.instance_type = instance_type
        self.vpc_id = vpc_id
        self.s3_bucket = s3_bucket  # needed for IAM setup; bucket will not be created by terraform
        self.job_dir = os.path.abspath(job_dir)

        self.image_dir = None
        self.ssh = None
        self.remote_workdir = '/home/ec2-user/image-atm'

        self.__check_s3_prefix()

        self.logger = get_logger(__name__, self.job_dir)

    def __check_s3_prefix(self):
        # ensure that s3 bucket prefix is correct
        self.s3_bucket_wo = self.s3_bucket.split('s3://')[-1]  # without s3:// prefix
        self.s3_bucket = 's3://' + self.s3_bucket_wo

    def __set_ssh(self):
        cmd = 'cd {} && terraform output public_ip'.format(self.tf_dir)
        output = run_cmd(cmd, logger=self.logger, return_output=True)
        self.ssh = 'ssh -o StrictHostKeyChecking=no -i ~/.ssh/id_rsa ec2-user@{}'.format(output)

    def __set_remote_dirs(self):
        cmd = '{} mkdir -p {}'.format(self.ssh, self.remote_workdir)
        run_cmd(cmd, logger=self.logger)

        self.remote_image_dir = os.path.join(self.remote_workdir, os.path.basename(self.image_dir))
        self.remote_job_dir = os.path.join(self.remote_workdir, os.path.basename(self.job_dir))

    def __set_s3_dirs(self):
        if 's3://' in self.image_dir:
            self.s3_image_dir = self.image_dir
        else:
            self.s3_image_dir = os.path.join(
                self.s3_bucket, 'image_dirs', os.path.basename(self.image_dir)
            )

        if 's3://' in self.job_dir:
            self.s3_job_dir = self.job_dir
        else:
            self.s3_job_dir = os.path.join(
                self.s3_bucket, 'job_dirs', os.path.basename(self.job_dir)
            )

    def __sync_local_s3(self):
        self.logger.info('Syncing files local <> s3...')
        self.__set_s3_dirs()

        # only sync if image dir is local dir
        if 's3://' not in self.image_dir:
            cmd = 'aws s3 sync --quiet {} {}'.format(self.image_dir, self.s3_image_dir)
            run_cmd(cmd, logger=self.logger)

        # only sync if job dir is local dir
        if 's3://' not in self.job_dir:
            cmd = 'aws s3 sync --quiet {} {}'.format(self.job_dir, self.s3_job_dir)
            run_cmd(cmd, logger=self.logger)

    def __sync_s3_local(self):
        self.logger.info('Syncing files s3 <> local...')
        self.__set_s3_dirs()

        # only sync if job dir is local dir
        if 's3://' not in self.job_dir:
            cmd = 'aws s3 sync --quiet {} {}'.format(self.s3_job_dir, self.job_dir)
            run_cmd(cmd, logger=self.logger, level='info')

    def __sync_remote_s3(self):
        self.logger.info('Syncing files remote <> s3...')
        self.__set_s3_dirs()
        self.__set_remote_dirs()

        cmd = '{} aws s3 sync --quiet {} {}'.format(
            self.ssh, self.remote_image_dir, self.s3_image_dir
        )
        run_cmd(cmd, logger=self.logger)

        cmd = '{} aws s3 sync --quiet {} {}'.format(self.ssh, self.remote_job_dir, self.s3_job_dir)
        run_cmd(cmd, logger=self.logger)

    def __sync_s3_remote(self):
        self.logger.info('Syncing files s3 <> remote...')
        self.__set_s3_dirs()
        self.__set_remote_dirs()

        cmd = '{} aws s3 sync --quiet {} {}'.format(
            self.ssh, self.s3_image_dir, self.remote_image_dir
        )
        run_cmd(cmd, logger=self.logger)

        cmd = '{} aws s3 sync --quiet {} {}'.format(self.ssh, self.s3_job_dir, self.remote_job_dir)
        run_cmd(cmd, logger=self.logger)

    def __launch_train_container(self, **kwargs):
        self.logger.info('Launching training container...')
        # training parameters are passed to container through environment variables
        envs = ['-e {}={}'.format(key, value) for key, value in kwargs.items()]
        cmd = (
            '{} docker run -d -v {}:$WORKDIR/image_dir -v {}:$WORKDIR/job_dir {} '
            'idealo/tensorflow-image-atm:1.12.0'
        ).format(self.ssh, self.remote_image_dir, self.remote_job_dir, ' '.join(envs))
        run_cmd(cmd, logger=self.logger)

    def __stream_docker_logs(self):
        time.sleep(5)
        cmd = '{} docker ps -l -q'.format(self.ssh)
        output = run_cmd(cmd, logger=self.logger, return_output=True)

        cmd = '{} docker logs {} --follow'.format(self.ssh, output)
        run_cmd(cmd, logger=self.logger, level='info')

    def init(self):
        self.logger.info('Running terraform init...')
        cmd = 'cd {} && terraform init'.format(self.tf_dir)
        run_cmd(cmd, logger=self.logger)

    def apply(self):
        self.logger.info('Running terraform apply...')
        cmd = (
            'cd {} && terraform apply -auto-approve -var "region={}" -var "instance_type={}" '
            '-var "vpc_id={}" -var "s3_bucket={}"'
        ).format(self.tf_dir, self.region, self.instance_type, self.vpc_id, self.s3_bucket_wo)

        run_cmd(cmd, logger=self.logger)

        self.__set_ssh()

    def train(self, image_dir=None, job_dir=None, **kwargs):
        self.logger.info('Setting up remote instance...')
        if image_dir is not None:
            self.image_dir = os.path.abspath(image_dir)

        self.__sync_local_s3()
        self.__sync_s3_remote()
        self.__launch_train_container(**kwargs)
        self.__stream_docker_logs()
        self.__sync_remote_s3()
        self.__sync_s3_local()

    def destroy(self):
        self.logger.info('Running terraform destroy...')
        cmd = (
            'cd {} && terraform destroy -auto-approve -var "region={}" -var "instance_type={}" '
            '-var "vpc_id={}" -var "s3_bucket={}"'
        ).format(self.tf_dir, self.region, self.instance_type, self.vpc_id, self.s3_bucket_wo)

        run_cmd(cmd, logger=self.logger)
